
// This is a filter. A filter takes in an item, and returns either:
// - an item
// - a promise for an item
// - a falsy value
// - a promise for a falsy value
// Items can be modified, kept the same, or filtered out altogether.

// Filters will be chained together by the soclial core, passed as parameters
// to itemPromise.then(...). This allows for both sync and async filters.

import urlRegexFn from 'url-regex';
const urlRegex = urlRegexFn();	// For some reason the module exports a function

// import { extract } from 'article-parser';
import Deshortifier from '../../node_modules/deshortify/src/deshortify'
// import Deshortifier from '../../../deshortify/src/deshortify'

// console.log(Deshortifier);

let deshortifier = new Deshortifier({ verbose: false });

// let cache = {};


export default function deshortify(item) {

	// This filter takes the text in the item, finds the URLs, and deshortifies them.

	let urls = item.str.match(urlRegex);

	if (!urls) { urls = []; }	// Item itself doesn't have any URLs, but might have an echo/quote

	item._deshortifyOriginalUrls = urls;
// 	console.log('Detected URLs:', urls);

	var ownUrlPromises = (Array.from(urls)).map( (url)=>{
		return deshortifier.deshortify(url)
			.then(deshortified=>[url, deshortified])
			.catch(err=>console.error(err));
	});

	let ownItemPromise = Promise.all(ownUrlPromises).then( resolvedUrls=>{

		item._deshortifyResolvedUrls = resolvedUrls;

		// 		console.log(resolvedUrls);
		resolvedUrls.forEach(([original, replacement])=>{
			if (original !== replacement) {
// 				console.log('Replacing: ', original, ' → ', replacement);
				item.str = item.str.replace(original, replacement);
			}
		});

		return item;
	});

	// Does this item echo another? Then deshortify its URLs too.
	if (item.echoed) {
//             console.log('Deshortifying also: ', item.echoed);
		let echoedItemPromise = deshortify(item.echoed);

		ownItemPromise = Promise.all([ownItemPromise, echoedItemPromise]).then(([own, echoed])=>{
			own.echoed = echoed;
			return own;
		});
	}

	// Does this item quote another? Then deshortify its URLs too.
	if (item.quoted) {
// 		console.log('Deshortifying also: ', item.quoted);
		let echoedItemPromise = deshortify(item.quoted);

		ownItemPromise = Promise.all([ownItemPromise, echoedItemPromise]).then(([own, quoted])=>{
			own.quoted = quoted;
			return own;
		});
	}


	return Promise.race([
		ownItemPromise,
		new Promise(function (resolve, reject) { setTimeout(reject, 10000) })
	]).catch((err=>{
		/// FIXME: Should be vorpal + chalk
		console.error('Resolving URLs took too long. URLs were:', urls);
		return Promise.resolve(item);
	}));

}







