
import Account from '../abstracts/account';
import TimeLock from '../timelock';

import inquirer from 'inquirer';
import { OAuth } from 'oauth';
import chalk from 'chalk256';
import Twitter from 'twitter';
// import striptags from 'striptags';
// import he from 'he';

const soclialConsumerKey = 'xiUNN9g7K8it8QIXcA';
const soclialConsumerSecret = 'qsFwbc6gXtcHAplAXjfoKXu96hVXz82W7sEbzuLR2U';

// 🍂class TwitterAccount
// 🍂inherits Account
// Support for the Twitter microblogging network.

class TwitterAccount extends Account {

	constructor(callback, vorpal, accountNumber, opts) {

		super.constructor(callback, opts);

// 		/// While we are debugging, it's a good idea to not ask too much to the twitter servers.
// 		return;	/// DEBUG!!!

		this._opts = Object.assign({
// 			colour: 'grey'
		}, opts);

		this._tw = new Twitter({
			consumer_key: soclialConsumerKey,
			consumer_secret: soclialConsumerSecret,
			access_token_key: opts.clientId,
			access_token_secret: opts.clientSecret
		});

		this._vorpal = vorpal;
		this._accountNumber = accountNumber;

		this._lenghtOfTCoLinks = 30;	// Start with a conservative value

		let initialTimeLock = new TimeLock();

		// Lock the stream until we have some data from all the channels
		this.emit(-Infinity, initialTimeLock.promise);
// 		console.log('TW platform locked the main stream on user init');

		Promise.all([
			this._getEndpointPromise('account/verify_credentials'),
			this._getEndpointPromise('application/rate_limit_status')
		]).then(([
			credentials,
			rateLimits
		])=>{

			vorpal.log(
				chalk.purple('You are logged in the Twitter platform as '),
				chalk.cyan(credentials.screen_name),
// 				chalk.purple(' in server '),
// 				chalk.cyan(opts.api_url.replace(/.*:\/\//,'').replace(/\/.*/, '')),
				chalk.purple(', account badge is '),
				chalk.cyan(opts.badge)
			);

			vorpal.log(
				chalk.purple('Rate limits at start-up time are:'),
				chalk.cyan(JSON.stringify(rateLimits))
			);


			this._id = credentials.screen_name + '@twitter';

			this._idStr = credentials.id_str;

			// Hackishly listen to the user's main stream
			// TODO: Split this out, set a flag when streaming is working
			// TODO: Handle errors when the network connectivity changes
			// TODO: Handle errors when receiving a HTTP status response of 420 (connecting too often, or too many clients)
			this._stream = this._tw.stream('user', {
// 				replies: "all",	// Should explicitly listen to the "replies" channel
// track: "foo,bar"	// Should use this for listening to search terms channels
				stall_warnings: true
			}, (stream)=>{
				stream.on('data', (ev)=>{
// 					vorpal.log(ev);
					this._normalizeAndEmitUpdate(ev, 'home');
				});

				stream.on('error', function(error) {
					vorpal.log(chalk.red(error));
				});

			});

// 			// Hackishly listen to the "samples" public timeline
// 			this._stream = this._tw.stream('statuses/sample', {
// // 				replies: "all",	// Should explicitly listen to the "replies" channel
// 				// track: "foo,bar"	// Should use this for listening to search terms channels
// 				stall_warnings: true
// 			}, (stream)=>{
// 				stream.on('data', (ev)=>{
// // 					vorpal.log(ev);
// 					this._normalizeAndEmitUpdate(ev, 'public');
// 				});
//
// 				stream.on('error', function(error) {
// 					vorpal.log(chalk.red(error));
// 				});
// 			});


			let homeTimelineLock = new TimeLock();
			this.emit(-Infinity, homeTimelineLock.promise);
// 			console.log('TW locked the main stream on the last home timeline');

			// Get a few of the last statuses in the main timeline
			this._tw.get('statuses/home_timeline', {
				count: 100,

				// This undocumented option makes the tweets have the full text,
				// and never be truncated.
				tweet_mode: 'extended'
// 				count: 10
			}, (err, msgs, response)=>{
				if (err) {
					vorpal.log(chalk.red(err));
				}

				for (let i=msgs.length-1; i>=0; i--) {
// 					console.log(msgs[i]);
					this._normalizeAndEmitUpdate(msgs[i], 'home');
// 					vorpal.log('timelines/home', msgs);
				}
// 				console.log('TW removed home timeline lock');
				homeTimelineLock.unlock();
			});

			initialTimeLock.unlock();
// 			console.log('TW platform unlocked user init');


			// Fetches some "constants" from the API.
			// See https://dev.twitter.com/rest/reference/get/help/configuration
			this._tw.get('help/configuration', {}, (err, json)=>{
				if (err) { return vorpal.log(chalk.red(err)); }
				this._lenghtOfTCoLinks = json.short_url_length_https;
				return vorpal.log(
					chalk.purple('>>> Length of t.co links in the twitter platform is currently: '),
					chalk.cyan(this._lenghtOfTCoLinks)
				);
			});

		});
	}

	listen(channel) {
/// FIXME!!!
	}


	unlisten(channel) {
/// FIXME!!!
	}

	listChannels() {
// 		return ['home', 'replies', 'notifications', 'public'];
	}

	getLastUpdateTimestamp() {
// 		return Date.now();
	}

	post(str) {
		return this._post({ status: str });
	}

	reply (item, str) {
		let replyStatusId = item.raw.id_str;
		let replyStatusScreenName = "@" + item.raw.user.screen_name;

		if (str.toLowerCase().indexOf( replyStatusScreenName.toLowerCase() ) === -1) {
			/// FIXME!!! This will match partial usernames (e.g. searching for @john
			/// will match @johnDoe). Should use some kind of regexp here.

			str = replyStatusScreenName + ' ' + str;
		}

		return this._post({
			status: str,
			in_reply_to_status_id: replyStatusId
		});
	}

	echoify(item) {
		if (item.echoed) {
			return "RT @" + item.raw.retweeted_status.user.screen_name + ': ' + item.echoed.str;
		}

		return "RT @" + item.raw.user.screen_name + ': ' + item.str;
	}

	echo(item, str) {
		if (item.platform === 'twitter') {
			// Native RT
			return this._tw.post('statuses/retweet', {
				id: item.raw.id_str
			},  (error, tweet, response)=>{
				if(error) {
					this._vorpal.log(chalk.red('Could not echo into twitter: ' + error[0].message));
				}
			});
		} else {
			/// TODO: Take into account that twitter shortens ALL URLs into t.co links,
			/// potentially shortening the length
			let totalLength = str.length + 1 + this._lenghtOfTCoLinks;
			if (totalLength > 140) {
				// Time to clip something, and replace rest with an ellipsis
				str = str.substr(0, 140 - this._lenghtOfTCoLinks - 2) + '…';
			}
			str += ' ' + item.url;

			return this.post(str);
		}
	}

	// Common to post, reply, replyall, echo
	_post(payload) {
		this._tw.post('statuses/update', payload,  function(error, tweet, response) {
			if(error) throw error;
// 			console.log(tweet);  // Tweet body.
// 			console.log(response);  // Raw response object.
		});
	}

	// A helper to promisify calls to this._tw.get().
	_getEndpointPromise(endpoint) {
		return new Promise((res, rej)=>{
			this._tw.get(endpoint, {}, (err, json)=>{
				return err ? rej(err) : res(json);
			});
		});
	}

	_cleanString(str, entities, inReplyTo) {
		if (entities.urls) {
			entities.urls.forEach(el=>str = str.replace(el.url, el.expanded_url));
		}
		if (entities.media) {
			// Media is handled by soCLIal and should not be included in the text
			entities.media.forEach(el=>str = str.replace(el.url, ''));
		}

		if (inReplyTo) {
			str = "[re: @" + inReplyTo + "] " + str;
		}

// 		return striptags( he.decode( str.replace('</p><p>', ' // ') ) );
		return str.
			replace(/&amp;/g, "&").
			replace(/&gt;/g, ">").
			replace(/&lt;/g, "<");
	}

	// Expand the media URLs. This is not trivial because of video URLs.
	_cleanMedia(msg) {
		let rawEntities = {};

		if (!msg.entities) { return }

		if (msg.entities && msg.entities.media) {
			msg.entities.media.forEach(el=>rawEntities[el.id_str]=el);
		}
		if (msg.extended_entities && msg.extended_entities.media) {
			msg.extended_entities.media.forEach(el=>rawEntities[el.id_str]=el);
		}

		return Object.keys(rawEntities).map(k=>{
			let ent = rawEntities[k];
			if ('video_info' in ent) {
				return ent.video_info.variants[0].url;
			}
			return ent.media_url.https || ent.media_url;
		});
	}

// 	// Given a stream descriptor, return a function that handles messages from the stream
// 	_getOnStreamMessage(stream) {
// 		return (msg)=>{
// // 			this._vorpal.log(chalk.green(msg));
// 			if (msg.event === 'update') {
// 				this._normalizeAndEmitUpdate(stream, msg.data);
// 			} else if (msg.event === 'notification') {
// 				this._normalizeAndEmitNotification(stream, msg.data);
// 			} else {
// 				this._vorpal.log(chalk.green(msg));
// 			}
// 		}
// 	}

	// Get a Twitter update, OR a part of an update (such a retweeted_status or a
	// quoted_status), and normalize some stuff (to make different platforms
	// see the same data from different platforms)
	_normalize(msg, channel) {
// 		if (('id_str' in msg) && ('contributors' in msg) && ('text' in msg)) {
		if (('id_str' in msg) && (
				('contributors' in msg) && ('text' in msg) ||	// Non-extended
				('full_text' in msg)	// Extended
		)) {
			// This looks like a tweet

// 			console.log('tweet is truncated: ', msg.truncated);
// 			console.log(msg.entities.urls);

// 			if (msg.truncated) {
// 				// This is the retweeted part of a RT update.
// 				/// TODO: Return a promise that fetches the original, untruncated tweet,
// 				/// and normalizes it.
//
// 			}

			if (msg.quoted_status &&
			    msg.retweeted_status &&
			    msg.retweeted_status.quoted_status &&
			    msg.retweeted_status.quoted_status.id_str === msg.quoted_status.id_str
			) {
				// Delete the quoted one, as the quoted from the retweet is the same
				// and has the extended attributes
				delete msg.quoted_status;
			}

			if (msg.extended_tweet && msg.extended_tweet.full_text) {
				msg.text = msg.extended_tweet.full_text;
				msg.truncated = false;
			}

			if (msg.truncated) {
// 				console.log('Truncated ', msg.text, ' - fetching full from id ', msg.id_str);
				return {
					timestamp: Date.parse(msg.created_at),
					itemPromise: this._tw.get('statuses/show/', {
						id: msg.id_str,
						tweet_mode: 'extended'
					}).then((full)=>{
						if (full.truncated) {
							// After requesting individual tweet, got it truncated also
	// 						console.log('Got full but still truncated. From: «', msg.text ,  '» into: ', full);
							console.log('Got full but still truncated. ', msg.text);
							full.truncated = false;
						}
						return this._normalize(full, channel).itemPromise;
					})
				}
			}

			if ((!msg.text) && msg.full_text) {
				// This looks like an extended tweet, fetched via a "tweet_mode=extended" REST API call
				msg.text = msg.full_text;
			}

			if (msg.extended_tweet) {
				// This looks like an extended tweet, fetched via a realtime stream
// 				return this._normalize(Object.assign({}, msg.extended_tweet, msg);
				msg.text = msg.extended_tweet.full_text;
				msg.entities = msg.extended_tweet.entities;
				/// TODO: Maybe something has to be done about the entities?
			}

			if (msg.retweeted_status) {
				// Wait for normalization of retweeted, return a promise
				return {
					timestamp: Date.parse(msg.created_at),
					itemPromise: this._normalize(msg.retweeted_status, null).itemPromise.then(echoed=>
						Promise.resolve({
							accountNumber: this._accountNumber,
							accountId: this._id,
							platform: 'twitter',
							raw: channel && msg,
							sender: msg.user.screen_name,
							str: '',
							url: 'https://www.twitter.com/' + msg.user.screen_name + '/status/' + msg.id_str,
							channel: channel,
							media: [],
							timestamp: Date.parse(msg.created_at),
							echoed: echoed
						})
					)
				};
// 					msg.text = msg.retweeted_status.text;
			}

			let allEntities = Object.assign(
				{},
				msg.entities || {},
				msg.extended_entities || {}
			);

// 			let allMedia = (allEntities.media && allEntities.media.map(i=>i.media_url_https || i.media_url )) || [];

// 			let allMedia = (allEntities.media && allEntities.media.map((i)=>{
// 				return
// // 					(i.video_info && i.video_info.variants[0].url) ||
// // 					i.expanded_url ||
// 					i.media_url_https ||
// 					i.media_url
// 			})) || [];

			let str = this._cleanString(msg.text, allEntities, msg.in_reply_to_screen_name);

			// Wait for normalization of quoted, return a promise
			if (msg.quoted_status) {
				return {
					timestamp: Date.parse(msg.created_at),
					itemPromise: this._normalize(msg.quoted_status, null).itemPromise.then(quoted=>
						Promise.resolve({
							accountNumber: this._accountNumber,
							accountId: this._id,
							platform: 'twitter',
							sender: msg.user.screen_name,
							str: str,
							url: 'https://www.twitter.com/' + msg.user.screen_name + '/status/' + msg.id_str,
							channel: channel,
							raw: channel && msg,
							media: this._cleanMedia(msg),
							timestamp: Date.parse(msg.created_at),
							quoted: quoted
						})
					)
				};
			}

			// Neither quoted nor retweeted here
			return {
				timestamp: Date.parse(msg.created_at),
				itemPromise: Promise.resolve({
					accountNumber: this._accountNumber,
					accountId: this._id,
					platform: 'twitter',
					sender: msg.user.screen_name,
					str: str,
					url: 'https://www.twitter.com/' + msg.user.screen_name + '/status/' + msg.id_str,
					channel: channel,
					raw: channel && msg,
					media: this._cleanMedia(msg),
					timestamp: Date.parse(msg.created_at),
				})
			};

		} else if (msg.direct_message) {

			let sender = msg.direct_message.sender_screen_name;
			if (msg.direct_message.sender_id === this._idStr) {
				sender = '→' + msg.direct_message.recipient_screen_name;
			}

			let str = this._cleanString(msg.direct_message.text, msg.direct_message.entities);

			return {
				timestamp: Date.parse(msg.direct_message.created_at),
				itemPromise: Promise.resolve({
					accountNumber: this._accountNumber,
					accountId: this._id,
					platform: 'twitter',
					sender: sender,
					str: str,
// 					url: 'https://www.twitter.com/' + msg.user.screen_name + '/status/' + msg.id_str,
					channel: 'DM',
					raw: channel && msg,
					media: this._cleanMedia(msg.direct_message),
					timestamp: Date.parse(msg.direct_message.created_at),
				})
			};

		} else {
			this._vorpal.log(chalk.red('Received something which is not a tweet nor a DM'));
			return {
				timestamp: Date.parse(msg.created_at),
				itemPromise:  Promise.resolve(msg)
			}
		}
	}



	// Get a Twitter update, normalizes some stuff (to make different platforms
	// see the same data from different platforms), and emits the update.
	_normalizeAndEmitUpdate(msg, channel) {
		let normalized = this._normalize(msg, channel);

		if (normalized) {
			this.emit(normalized.timestamp, normalized.itemPromise);
		}

// 		/// FIXME!!!
// 		if (msg.spoiler_text) {
// 			this._vorpal.log('Spoiler: ', msg.spoiler_text);
// 		}
//
// 		if (msg.retweeted_status) {
// 			this._vorpal.log('FIXME: This is a RT/echo of another status (truncated: ', msg.truncated, ')');
// // 			this._normalizeAndEmitUpdate(channel, msg.reblog);
// 		}
// 		if (msg.quoted_status) {
// 			this._vorpal.log('FIXME: This is a quote of another status');
// // 			this._normalizeAndEmitUpdate(channel, msg.reblog);
// 		}
	}


	_normalizeAndEmitNotification(msg) {
// 		if (msg.type === 'follow') {
// 			this.emit({
// 				str: chalk.bold(msg.account.acct) + ' started following you.',
// 				channel: channel,
// 				timestamp: Date.parse(msg.created_at)
// 			});
// 		} else if (msg.type === 'reblog') {
// 			this.emit({
// 				str: chalk.bold(msg.account.acct) + ' echoed your update: ' + this._cleanString(msg.status.content),
// 				channel: channel,
// 				timestamp: Date.parse(msg.created_at)
// 			});
// 		} else if (msg.type === 'favourite') {
// 			this.emit({
// 				str: chalk.bold(msg.account.acct) + ' favourited your update: ' + this._cleanString(msg.status.content),
// 				channel: channel,
// 				timestamp: Date.parse(msg.created_at)
// 			});
// 		} else if (msg.type === 'mention') {
// 			this.emit({
// 				str: chalk.bold(msg.account.acct) + ' mentioned you ' + this._cleanString(msg.status.content),
// 				channel: channel,
// 				timestamp: Date.parse(msg.created_at)
// 			});
// 		} else {
// 			this._vorpal.log('channel', msg);
// 		}
	}

}


// signUp returns a Promise for a set of options. If the promise resolves, these options
// will be saved into the user preferences, then a new instance of this platform's `Account`
// will be instantiated.
export function signUp() {

	return new Promise( (resolve, reject) => {
		/// Twitter only has one known API endpoint, no need to handle this.
		let server = 'api.twitter.com';

		let oauth = new OAuth(
			"https://" + server + "/oauth/request_token",
			"https://" + server + "/oauth/access_token",
			soclialConsumerKey,
			soclialConsumerSecret,
			"1.0A",
			'oob',
			"HMAC-SHA1"
		);

// 		console.log('oauth instance: ', oauth);

		oauth.getOAuthRequestToken( (error, oauthToken, oauthTokenSecret, oauthAuthorizeUrl, additionalParameters )=>{
			if(error) {
				return reject( 'Error retrieving the OAuth Request Token: ' + JSON.stringify(error) );
			}

// 			console.log('Successfully retrieved the OAuth Request Token:', oauthToken,
// 						', secret: ', oauthTokenSecret,
// 						', auth url:', oauthAuthorizeUrl);

			let url = "http://" + server + "/oauth/authenticate?oauth_token=" + oauthToken;
			console.log( chalk.bold('In order to add a new Twitter account, visit ', chalk.underline(url), ' to get a verification code.') );

			return inquirer.prompt([{
				type: 'input',
				name: 'code',
				message: 'Enter verification code: '
			}]).then((answer)=>{
				if (!answer.code) {
					return reject('No verification code was introduced.');
				}

// 				console.log('We got verification code:', answer.code);

				oauth.getOAuthAccessToken(
					oauthToken,
					oauthTokenSecret,
					answer.code,
					function( error, oauthClientToken, oauthClientSecret, additionalParameters ) {
						if( error ) {
							return reject( 'Error retrieving the OAuth Client Token: ' + JSON.stringify(error) );
						}

						return resolve({
							clientId: oauthClientToken,
							clientSecret: oauthClientSecret
						});
					}
				);
			});
		});
	});

}

export const defaultBadge = 'tw';
export const name = 'twitter';
export { TwitterAccount as Account };

